#![feature(portable_simd)]
#![feature(stdarch_arm_neon_intrinsics)]

use anyhow::{anyhow, Context, Result};
use clap::{Parser, ValueEnum};
use std::{
    fs::File,
    io::{BufRead, BufReader, Read, Write},
    net::{TcpListener, TcpStream},
    process::Command,
    time::Duration,
};

mod reduce;
mod rmframes;
use rmframes::RMFrames;

// Parameters only for RM2, firmware version >= 3.7.0.1930

/// Height (in pixels) of the framebuffer.
const HEIGHT: usize = 1404;
/// Width (in pixels) of the framebuffer.
const WIDTH: usize = 1872;
/// How many bytes represent one pixel in the framebuffer.
const BYTES_PER_PIXEL: usize = 2;

const IN_SIZE: usize = WIDTH * HEIGHT * BYTES_PER_PIXEL;

#[derive(Parser, Debug)]
#[command(author, version)]
pub struct Opts {
    #[arg(long, name = "port", short = 'p')]
    port: usize,

    #[arg(long, name = "format", short = 'f')]
    format: Format,
}

#[derive(ValueEnum, Debug, Clone, Copy)]
pub enum Format {
    Mono,
    MonoSimd,
    Gray8,
    Gray8Simd,
    Gray16be,
}

impl Format {
    fn run<W: Write>(&self, mut src: RMFrames, mut tgt: W) -> Result<()> {
        const IN_SIZE: usize = WIDTH * HEIGHT * BYTES_PER_PIXEL;

        match self {
            Self::Gray16be => {
                let mut buf = [0u8; IN_SIZE];
                loop {
                    src.read_exact(&mut buf)?;
                    tgt.write_all(&buf[0..IN_SIZE])?;
                }
            }

            Self::Gray8 => {
                const OUT: usize = reduce::gray8::out_size(IN_SIZE);
                let mut buf = [0u8; IN_SIZE];
                loop {
                    src.read_exact(&mut buf)?;
                    reduce::gray8::run(&mut buf);
                    tgt.write_all(&buf[0..OUT])?;
                }
            }

            Self::Gray8Simd => {
                const OUT: usize = reduce::gray8_simd::out_size(IN_SIZE);
                let mut buf = [0u8; IN_SIZE];
                loop {
                    src.read_exact(&mut buf)?;
                    reduce::gray8_simd::run(&mut buf);
                    tgt.write_all(&buf[0..OUT])?;
                }
            }

            Self::Mono => {
                const OUT: usize = reduce::mono::out_size(IN_SIZE);
                let mut buf = [0u8; IN_SIZE];
                loop {
                    src.read_exact(&mut buf)?;
                    reduce::mono::run(&mut buf);
                    tgt.write_all(&buf[0..OUT])?;
                }
            }

            Self::MonoSimd => {
                const OUT: usize = reduce::mono_simd::out_size(IN_SIZE);
                let mut buf = [0u8; IN_SIZE];
                loop {
                    src.read_exact(&mut buf)?;
                    reduce::mono_simd::run(&mut buf);
                    tgt.write_all(&buf[0..OUT])?;
                }
            }
        }
    }
}

fn main() -> Result<()> {
    let opts: Opts = Opts::parse();

    let (file, offset) = {
        let pid = xochitl_pid()?;
        let offset = rm2_fb_offset(pid)?;
        let mem = format!("/proc/{}/mem", pid);
        (mem, offset)
    };

    let src = RMFrames::init(&file, offset, IN_SIZE)?;
    let tgt = listen_timeout(opts.port, Duration::from_secs(60))?;

    opts.format.run(src, tgt)?;

    Ok(())
}

fn listen_timeout(port: usize, timeout: Duration) -> Result<TcpStream> {
    let listen_addr = format!("0.0.0.0:{}", port);
    let listen = TcpListener::bind(&listen_addr)?;
    eprintln!("[rM] listening for a TCP connection on {}", listen_addr);

    let (conn, conn_addr) = listen.accept().unwrap();

    eprintln!("[rM] connection received from {}", conn_addr);
    conn.set_write_timeout(Some(timeout))?;
    Ok(conn)
}

fn xochitl_pid() -> Result<usize> {
    let output = Command::new("/bin/pidof")
        .args(["xochitl"])
        .output()
        .context("Failed to run `/bin/pidof xochitl`")?;
    if output.status.success() {
        let pid = &output.stdout;
        let pid_str = std::str::from_utf8(pid)?.trim();
        pid_str
            .parse()
            .with_context(|| format!("Failed to parse xochitl's pid: {}", pid_str))
    } else {
        Err(anyhow!(
            "Could not find pid of xochitl, is xochitl running?"
        ))
    }
}

fn rm2_fb_offset(pid: usize) -> Result<usize> {
    let file = File::open(format!("/proc/{}/maps", &pid))?;
    let line = BufReader::new(file)
        .lines()
        .skip_while(|line| matches!(line, Ok(l) if !l.ends_with("/dev/fb0")))
        .nth(1)
        .with_context(|| format!("No line containing /dev/fb0 in /proc/{}/maps file", pid))?
        .with_context(|| format!("Error reading file /proc/{}/maps", pid))?;

    let addr = line
        .split("-")
        .next()
        .with_context(|| format!("Error parsing line in /proc/{}/maps", pid))?;

    let address = usize::from_str_radix(addr, 16).context("Error parsing framebuffer address")?;
    Ok(address + 8)
}
